# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

from typing import List

import pabot.pabot


def run_pabot(
    path: str, include: List[str] = [], exclude: List[str] = [], dry_run: bool = False
) -> None:
    """Run pabot"""
    args = ["--pabotlib"]
    if dry_run:
        args.append("--dryrun")
    for i in include:
        args.extend(["--include", i])
    for e in exclude:
        args.extend(["--exclude", e])
    args.extend(
        ["-d", path, "--skiponfailure", "non-critical", "-x", "xunit.xml", path]
    )
    pabot.pabot.main(args)



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import importlib.util
import logging
import os
import subprocess
from typing import Any, Dict, List

from ruamel import yaml

logger = logging.getLogger(__name__)


class VaultTag(yaml.YAMLObject):
    yaml_tag = "!vault"

    def __init__(self, v: str):
        self.value = v

    def __repr__(self) -> str:
        spec = importlib.util.find_spec("iac_test.ansible_vault")
        if spec:
            if "ANSIBLE_VAULT_ID" in os.environ:
                vault_id = os.environ["ANSIBLE_VAULT_ID"] + "@" + str(spec.origin)
            else:
                vault_id = str(spec.origin)
            t = subprocess.check_output(
                [
                    "ansible-vault",
                    "decrypt",
                    "--vault-id",
                    vault_id,
                ],
                input=self.value.encode(),
            )
            return t.decode()
        return ""

    @classmethod
    def from_yaml(cls, loader: Any, node: Any) -> str:
        return str(cls(node.value))


class EnvTag(yaml.YAMLObject):
    yaml_tag = "!env"

    def __init__(self, v: str):
        self.value = v

    def __repr__(self) -> str:
        env = os.getenv(self.value)
        if env is None:
            return ""
        return env

    @classmethod
    def from_yaml(cls, loader: Any, node: Any) -> str:
        return str(cls(node.value))


def load_yaml_files(paths: List[str]) -> Dict[str, Any]:
    """Load all yaml files from a provided directory."""

    def _load_file(file_path: str, data: Dict[str, Any]) -> None:
        with open(file_path, "r") as file:
            if ".yaml" in file_path or ".yml" in file_path:
                data_yaml = file.read()
                y = yaml.YAML()
                y.preserve_quotes = True  # type: ignore
                y.register_class(VaultTag)
                y.register_class(EnvTag)
                dict = y.load(data_yaml)
                merge_dict(dict, data)

    result: Dict[str, Any] = {}
    for path in paths:
        if os.path.isfile(path):
            _load_file(path, result)
        else:
            for dir, subdir, files in os.walk(path):
                for filename in files:
                    try:
                        _load_file(dir + os.path.sep + filename, result)
                    except:  # noqa: E722
                        logger.warning("Could not load file: {}".format(filename))
    return result


def merge_list_item(
    source_item: Any, destination: List[Any], merge_list_items: bool = True
) -> None:
    """Merge item into list."""
    if isinstance(source_item, dict) and merge_list_items:
        # check if we have an item in destination with matching primitives
        for dest_item in destination:
            match = True
            comparison = False
            unique_source = False
            unique_dest = False
            for k, v in source_item.items():
                if isinstance(v, dict) or isinstance(v, list):
                    continue
                if k in dest_item and v == dest_item[k]:
                    comparison = True
                    continue
                if k not in dest_item:
                    unique_source = True
                    continue
                comparison = True
                match = False
            for k, v in dest_item.items():
                if isinstance(v, dict) or isinstance(v, list):
                    continue
                if k in source_item and v == source_item[k]:
                    comparison = True
                    continue
                if k not in source_item:
                    unique_dest = True
                    continue
                comparison = True
                match = False
            if comparison and match and not (unique_source and unique_dest):
                merge_dict(source_item, dest_item, merge_list_items)
                return
    elif source_item in destination:
        return
    destination.append(source_item)


def merge_dict(
    source: Dict[Any, Any], destination: Dict[Any, Any], merge_list_items: bool = True
) -> Dict[Any, Any]:
    """Merge two nested dict/list structures."""
    if not source:
        return destination
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            if node is None:
                destination[key] = value
            else:
                merge_dict(value, node, merge_list_items)
        elif isinstance(value, list):
            if key not in destination:
                destination[key] = []
            if isinstance(destination[key], list):
                for i in value:
                    merge_list_item(i, destination[key], merge_list_items)
        else:
            destination[key] = value
    return destination



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import click

data = click.option(
    "-d",
    "--data",
    type=click.Path(exists=True, dir_okay=True, file_okay=True),
    envvar="IAC_TEST_DATA",
    help="Path to data YAML files. (env: IAC_TEST_DATA)",
    required=True,
    multiple=True,
)

templates = click.option(
    "-t",
    "--templates",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    envvar="IAC_TEST_TEMPLATES",
    help="Path to test templates. (env: IAC_TEST_TEMPLATES)",
    required=True,
)

filters = click.option(
    "-f",
    "--filters",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    envvar="IAC_TEST_FILTERS",
    help="Path to Jinja filters. (env: IAC_TEST_FILTERS)",
    required=False,
)

tests = click.option(
    "--tests",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    envvar="IAC_TEST_TESTS",
    help="Path to Jinja tests. (env: IAC_TEST_TESTS)",
    required=False,
)


output = click.option(
    "-o",
    "--output",
    type=click.Path(exists=False, dir_okay=True, file_okay=False),
    envvar="IAC_TEST_OUTPUT",
    help="Path to output directory. (env: IAC_TEST_OUTPUT)",
    required=True,
)

include = click.option(
    "-i",
    "--include",
    envvar="IAC_TEST_INCLUDE",
    help="Selects the test cases by tag (include). (env: IAC_TEST_INCLUDE)",
    required=False,
    multiple=True,
)

exclude = click.option(
    "-e",
    "--exclude",
    envvar="IAC_TEST_EXCLUDE",
    help="Selects the test cases by tag (exclude). (env: IAC_TEST_EXCLUDE)",
    required=False,
    multiple=True,
)

render_only = click.option(
    "--render-only",
    is_flag=True,
    envvar="IAC_TEST_RENDER_ONLY",
    help="Only render tests without executing them. (env: IAC_TEST_RENDER_ONLY)",
)

dry_run = click.option(
    "--dry-run",
    is_flag=True,
    envvar="IAC_DRY_RUN",
    help="Dry run flag. See robot dry run mode. (env: IAC_DRY_RUN)",
    required=False,
)



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import os
import sys

if "ANSIBLE_VAULT_PASSWORD" in os.environ:
    sys.stdout.write("%s\n" % os.environ["ANSIBLE_VAULT_PASSWORD"])



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import logging
import sys
from typing import List

import click
import errorhandler

import iac_test.pabot
import iac_test.robot_writer

from . import options

logger = logging.getLogger(__name__)

error_handler = errorhandler.ErrorHandler()


def configure_logging(level: str) -> None:
    if level == "DEBUG":
        lev = logging.DEBUG
    elif level == "INFO":
        lev = logging.INFO
    elif level == "WARNING":
        lev = logging.WARNING
    elif level == "ERROR":
        lev = logging.ERROR
    else:
        lev = logging.CRITICAL
    logger = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(lev)
    error_handler.reset()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(iac_test.__version__)
@click.option(
    "-v",
    "--verbosity",
    metavar="LVL",
    is_eager=True,
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    help="Either CRITICAL, ERROR, WARNING, INFO or DEBUG",
    default="WARNING",
)
@options.data
@options.templates
@options.filters
@options.tests
@options.output
@options.include
@options.exclude
@options.render_only
@options.dry_run
def main(
    verbosity: str,
    data: List[str],
    templates: str,
    filters: str,
    tests: str,
    output: str,
    include: List[str],
    exclude: List[str],
    render_only: bool,
    dry_run: bool,
) -> None:
    """A CLI tool to render and execute Robot Framework tests using Jinja templating."""
    configure_logging(verbosity)

    writer = iac_test.robot_writer.RobotWriter(data, filters, tests, include, exclude)
    writer.write(templates, output)
    if not render_only:
        iac_test.pabot.run_pabot(output, include, exclude, dry_run)
    exit()


def exit() -> None:
    if error_handler.fired:
        sys.exit(1)
    else:
        sys.exit(0)



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import importlib.util
import json
import logging
import os
import pathlib
import re
import shutil
import sys
from typing import Any, Dict, List

from jinja2 import ChainableUndefined, Environment, FileSystemLoader  # type: ignore

from iac_test import yaml

logger = logging.getLogger(__name__)


class RobotWriter:
    def __init__(
        self,
        data_paths: List[str],
        filters_path: str,
        tests_path: str,
        include_tags: List[str] = [],
        exclude_tags: List[str] = [],
    ) -> None:
        logger.info("Loading yaml files from %s", data_paths)
        self.data = yaml.load_yaml_files(data_paths)
        self.filters: Dict[str, Any] = {}
        self.include_tags = include_tags
        self.exclude_tags = exclude_tags
        if filters_path:
            logger.info("Loading filters")
            for filename in os.listdir(filters_path):
                if filename.endswith(".py"):
                    file_path = os.path.join(filters_path, filename)
                    spec = importlib.util.spec_from_file_location(
                        "iac_test.filters", file_path
                    )
                    if spec is not None:
                        mod = importlib.util.module_from_spec(spec)
                        sys.modules["iac_test.filters"] = mod
                        if spec.loader is not None:
                            spec.loader.exec_module(mod)
                            self.filters[mod.Filter.name] = mod.Filter
        self.tests: Dict[str, Any] = {}
        if tests_path:
            logger.info("Loading tests")
            for filename in os.listdir(tests_path):
                if filename.endswith(".py"):
                    file_path = os.path.join(tests_path, filename)
                    spec = importlib.util.spec_from_file_location(
                        "iac_test.tests", file_path
                    )
                    if spec is not None:
                        mod = importlib.util.module_from_spec(spec)
                        sys.modules["iac_test.tests"] = mod
                        if spec.loader is not None:
                            spec.loader.exec_module(mod)
                            self.tests[mod.Test.name] = mod.Test

    def render_template(
        self, template_path: str, output_path: str, env: Environment, **kwargs: Any
    ) -> None:
        """Render single robot jinja template"""
        logger.info("Render robot template: %s", template_path)
        # add robot tags to kwargs
        kwargs["robot_include_tags"] = self.include_tags
        kwargs["robot_exclude_tags"] = self.exclude_tags
        # create output directory if it does not exist yet
        pathlib.Path(os.path.dirname(output_path)).mkdir(parents=True, exist_ok=True)

        template = env.get_template(template_path)
        # hack to convert nested ordereddict to dict, to avoid duplicate dict keys, e.g. 'tag'
        # json roundtrip should be safe as everything should be serializable
        data = json.loads(json.dumps(self.data))
        result = template.render(data, **kwargs)

        # remove extra empty lines
        lines = result.splitlines()
        cleaned_lines = []
        for index, line in enumerate(lines):
            if len(line.strip()):
                cleaned_lines.append(line)
            else:
                if index + 1 < len(lines):
                    next_line = lines[index + 1]
                    if len(next_line) and not next_line[0].isspace():
                        cleaned_lines.append(line)
        result = os.linesep.join(cleaned_lines)

        with open(output_path, "w") as file:
            file.write(result)

    def _fix_duplicate_path(self, *paths: str) -> str:
        """Helper function to detect existing paths with non-matching case. Returns a unique path to work with case-insensitve filesystems."""
        directory = os.path.join(*paths[:-1])
        if os.path.exists(directory):
            entries = os.listdir(directory)
            lower_case_entries = [path.lower() for path in entries]
            if paths[-1].lower() in lower_case_entries and paths[-1] not in entries:
                return os.path.join(*paths[:-1], "_" + paths[-1])
        return os.path.join(*paths)

    def write(self, templates_path: str, output_path: str) -> None:
        """Render Robot test suites."""
        env = Environment(
            loader=FileSystemLoader(templates_path),
            undefined=ChainableUndefined,
            lstrip_blocks=True,
            trim_blocks=True,
        )
        for name, filter in self.filters.items():
            env.filters[name] = filter.filter
        for name, test in self.tests.items():
            env.tests[name] = test.test

        for dir, _, files in os.walk(templates_path):
            for filename in files:
                if (
                    ".robot" not in filename
                    and ".j2" not in filename
                    and ".resource" not in filename
                ):
                    logger.info(
                        "Skip file with unknown file extension (not one of .robot, .resource or .j2): %s",
                        os.path.join(dir, filename),
                    )
                    out = os.path.join(
                        output_path, os.path.relpath(dir, templates_path)
                    )
                    pathlib.Path(out).mkdir(parents=True, exist_ok=True)
                    shutil.copy(os.path.join(dir, filename), out)
                    continue
                rel = os.path.relpath(dir, templates_path)
                t_path = os.path.join(rel, filename)

                # search for directives
                pattern = re.compile("{#(.+?)#}")
                content = ""
                next_template = False
                try:
                    with open(os.path.join(dir, filename), "r") as file:
                        content = file.read()
                except:  # noqa: E722
                    logger.warning(
                        "Could not open/read file: %s", os.path.join(dir, filename)
                    )
                    continue
                for match in re.finditer(pattern, content):
                    params = match.group().split(" ")
                    if len(params) == 6 and params[1] in [
                        "iterate_list",
                        "iterate_list_folder",
                    ]:
                        next_template = True
                        path = params[2].split(".")
                        attr = params[3]
                        elem = self.data
                        for p in path:
                            elem = elem.get(p, {})
                        if not isinstance(elem, list):
                            continue
                        for item in elem:
                            value = str(item.get(attr))
                            if value is None:
                                continue
                            extra = {}
                            if "[" in params[4]:
                                index = params[4].split("[")[1].split("]")[0]
                                extra_list = [None] * (int(index) + 1)
                                extra_list[int(index)] = value
                                extra = {params[4].split("[")[0]: extra_list}
                            else:
                                extra = {params[4]: value}
                            if params[1] == "iterate_list":
                                o_dir = self._fix_duplicate_path(
                                    output_path, rel, value
                                )
                                o_path = os.path.join(o_dir, filename)
                            else:
                                foldername = os.path.splitext(filename)[0]
                                new_filename = (
                                    value + "." + os.path.splitext(filename)[1][1:]
                                )
                                o_path = self._fix_duplicate_path(
                                    output_path, rel, foldername, new_filename
                                )
                            self.render_template(t_path, o_path, env, **extra)
                if next_template:
                    continue

                o_path = os.path.join(output_path, rel, filename)
                self.render_template(t_path, o_path, env)
